use std::time::Instant;

const BLOCK_SIZE: usize = 3;
const SIZE: usize = BLOCK_SIZE * BLOCK_SIZE;
const NUM_FIELDS: usize = SIZE * SIZE;

fn print_gray() {
    print!("\x1b\x5b\x31\x3b\x33\x30\x6d");
}

fn print_green() {
    print!("\x1b\x5b\x31\x3b\x33\x32\x6d");
}

fn print_neutral() {
    print!("\x1b\x5b\x31\x3b\x30\x6d");
}

fn print_clear() {
    print!("\x1b\x5b\x48\x1b\x5b\x32\x4a");
}

struct SField {
    field: Vec<u8>,
    skip_forward: Vec<u8>,
    skip_backward: Vec<u8>,
    pos: usize,
    possible_values: Vec<Vec<u8>>,
}

impl SField {
    pub fn new() -> SField {
        #[rustfmt::skip]
        let field = vec![
            0,0,0,0,0,0,0,0,0,
            0,0,0,0,8,9,0,0,0,
            0,0,0,0,2,0,4,0,0,
            0,0,4,0,6,0,0,0,8,
            0,0,0,5,0,0,0,0,0,
            0,6,5,0,0,2,0,7,4,
            0,3,0,0,0,5,0,4,0,
            0,0,1,8,0,0,0,0,0,
            0,0,8,2,0,0,0,6,0,
        ];

        let mut sudoku = SField {
            field,
            skip_forward: vec![0; NUM_FIELDS + 1],
            skip_backward: vec![0; NUM_FIELDS + 1],
            pos: 0,
            possible_values: vec![vec![]; NUM_FIELDS],
        };

        sudoku.build_possible_values_db();

        sudoku.build_skip_dbs();

        sudoku
    }

    fn build_skip_dbs(&mut self) {
        fn find_fixed_streak_forward(mut idx: usize, field: &[u8]) -> u8 {
            let mut fixed_count = 1;
            idx += 1;

            while idx < NUM_FIELDS && field[idx] > 0 {
                fixed_count += 1;
                idx += 1;
            }

            fixed_count
        }

        fn find_fixed_streak_backward(mut idx: usize, field: &[u8]) -> u8 {
            let mut fixed_count = 1;
            idx -= 1;

            while idx < NUM_FIELDS && field[idx] > 0 {
                fixed_count += 1;
                idx -= 1;
            }

            fixed_count
        }

        let fwd = &mut self.skip_forward;
        for (idx, nr) in self.field.iter().enumerate() {
            match nr {
                0 => fwd[idx] = *nr,
                _ => fwd[idx] = find_fixed_streak_forward(idx, &self.field),
            }
        }

        let bwd = &mut self.skip_backward;
        for (idx, nr) in self.field.iter().enumerate().rev() {
            match nr {
                0 => bwd[idx + 1] = *nr,
                _ => bwd[idx + 1] = find_fixed_streak_backward(idx, &self.field),
            }
        }
    }

    fn build_possible_values_db(&mut self) {
        for idx in 0..NUM_FIELDS {
            if self.field[idx] != 0 {
                continue;
            }
            self.pos = idx;
            // try all values between 1 and =self.size and remember the good ones
            let mut good_ones = Vec::with_capacity(SIZE);
            for nr in 1..=(u8::try_from(SIZE).expect("SIZE is too big")) {
                if self.ok(nr) {
                    good_ones.push(nr);
                }
            }
            self.possible_values[idx] = good_ones;
        }
        self.pos = 0;
    }

    pub fn solve_backtracking(&mut self) -> bool {
        let mut num_solutions = 0;

        loop {
            if !self.put_valid_nr() {
                self.clear_current_field();
                if !self.prev() {
                    print_clear();
                    self.print();
                    println!("Number of solutions: {num_solutions}");
                    break;
                }
                continue;
            }

            if !self.next() {
                num_solutions += 1;
                if num_solutions % 10_000 == 0 {
                    print_clear();
                    self.print();
                    println!("Number of solutions: {num_solutions}");
                }

                self.clear_current_field();
                self.prev();
            }
        }

        num_solutions > 0
    }

    fn print(&self) {
        for i in 0..NUM_FIELDS {
            if i != 0 && i % SIZE == 0 {
                println!();
            }

            if i == self.pos {
                print_green();
            } else if self.get_field_at_pos(i) == 0 {
                print_gray();
            }

            print!("{:2} ", self.get_field_at_pos(i));

            if i == self.pos || self.get_field_at_pos(i) == 0 {
                print_neutral();
            }
        }
        println!();
    }

    fn clear_current_field(&mut self) {
        self.set(0);
    }

    fn put_valid_nr(&mut self) -> bool {
        let current_nr = self.get_field_at_pos(self.pos);

        // safety:
        //     self.possible_vals is initialized with self.size * self.size
        //     so self.pos can safely be used to index here
        let possible_vals = &self.possible_values[self.pos];

        for nr in possible_vals {
            if *nr <= current_nr {
                continue;
            }

            if self.ok(*nr) {
                self.set(*nr);
                return true;
            }
        }
        false
    }

    fn ok(&self, nr: u8) -> bool {
        self.block_ok(nr) && self.row_ok(nr) && self.col_ok(nr)
    }

    fn get_field_at_pos(&self, pos: usize) -> u8 {
        // safety:
        //     TODO
        //     would need to mathematically explain that the calculations in
        //     get_row(), get_col() and get_block() can never exceed 0..self.num_fields
        unsafe { *self.field.get_unchecked(pos) }
    }

    fn set(&mut self, nr: u8) {
        self.field[self.pos] = nr;
    }

    fn get_row(&self, row: &mut [u8; SIZE]) {
        for (idx, row_elem) in row.iter_mut().enumerate() {
            *row_elem = self.get_field_at_pos((self.pos / SIZE) * SIZE + idx);
        }
    }

    fn get_col(&self, col: &mut [u8; SIZE]) {
        for (idx, col_elem) in col.iter_mut().enumerate() {
            *col_elem = self.get_field_at_pos(idx * SIZE + self.pos % SIZE);
        }
    }

    fn get_block(&self, block: &mut [u8; SIZE]) {
        let block_start_row = self.pos / SIZE / BLOCK_SIZE * BLOCK_SIZE;
        let block_start_col = self.pos % SIZE / BLOCK_SIZE * BLOCK_SIZE;

        for r in 0..BLOCK_SIZE {
            for c in 0..BLOCK_SIZE {
                block[r * BLOCK_SIZE + c] =
                    self.get_field_at_pos(SIZE * (block_start_row + r) + block_start_col + c);
            }
        }
    }

    fn block_ok(&self, nr: u8) -> bool {
        let mut block = [0; SIZE];
        self.get_block(&mut block);
        !block.contains(&nr)
    }

    fn row_ok(&self, nr: u8) -> bool {
        let mut row = [0; SIZE];
        self.get_row(&mut row);
        !row.contains(&nr)
    }

    fn col_ok(&self, nr: u8) -> bool {
        let mut col = [0; SIZE];
        self.get_col(&mut col);
        !col.contains(&nr)
    }

    fn next(&mut self) -> bool {
        let new_pos = self.pos + 1 + self.skip_forward[self.pos + 1] as usize;

        if new_pos >= NUM_FIELDS {
            return false;
        }

        self.pos = new_pos;
        true
    }

    fn prev(&mut self) -> bool {
        let new_pos = self.pos - 1 - self.skip_backward[self.pos] as usize;

        if new_pos >= NUM_FIELDS {
            return false;
        }

        self.pos = new_pos;
        true
    }

    fn solve(&mut self) {
        let now = Instant::now();
        if !self.solve_backtracking() {
            println!("there is no solution.");
        }
        eprintln!("took {:.3}s", now.elapsed().as_secs_f64());
    }
}

fn main() {
    let mut field = SField::new();
    field.solve();
}
